#include "adconfig.h"
#include "appframe.h"
#include "appframedialogimpl.h"
#include "toplevel.h"

#include <qwidget.h>
#include <qstring.h>
#include <qpixmap.h>
#include <qpushbutton.h>
#include <qtimer.h>
#include <qtooltip.h>
#include <qpainter.h>
#include <qpopupmenu.h>
#include <qdrawutil.h>

#include <kwin.h>
#include <kiconloader.h>
#include <klocale.h>
#include <kwinmodule.h>
#include <kdebug.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xmu/WinUtil.h>
#include <X11/cursorfont.h>


EmbedFrame::EmbedFrame( KWinModule *module,
    QWidget *parent, const char* name )
    : QWidget( parent, name ), winModule(module)
{
    _embeddedWinId = 0;

    _iconPixmap = new QPixmap();
    _bgPixmap = new QPixmap();

    XSelectInput( qt_xdisplay(), winId(),
        KeyPressMask | KeyReleaseMask |
        ButtonPressMask | ButtonReleaseMask |
        KeymapStateMask |
        ButtonMotionMask |
        PointerMotionMask |
        EnterWindowMask | LeaveWindowMask |
        FocusChangeMask |
        ExposureMask |
        StructureNotifyMask |
        SubstructureRedirectMask |
        SubstructureNotifyMask );

}

void EmbedFrame::setIconPixmap( QPixmap *pm )
{
    _iconPixmap = pm;
    repaint();
}

void EmbedFrame::setBgPixmap( QPixmap *pm )
{
    _bgPixmap = pm;
    repaint();
}

void EmbedFrame::embed( WId id )
{
    if( id == _embeddedWinId )
        return;

    QRect geom = KWin::info(id).geometry;

    // does the same as KWM::prepareForSwallowing()
    XWithdrawWindow( qt_xdisplay(), id, qt_xscreen() );
    while( KWin::info(id).state != WithdrawnState );

    // place application in the middle of the app frame
    if( (geom.width() < width()) ||
        (geom.height() < height()) )
        XReparentWindow( qt_xdisplay(), id, winId(),
            (width()-geom.width())/2,
            (height()-geom.height())/2 );
    else
        XReparentWindow( qt_xdisplay(), id, winId(), 0, 0 );

    // only resize if window is bigger than app frame
    if( (geom.width() > width()) ||
        (geom.height() > height()) )
        XResizeWindow( qt_xdisplay(), id, width(), height() );

    XMapWindow( qt_xdisplay(), id );
    XUngrabButton(
        qt_xdisplay(), AnyButton, AnyModifier, winId() );

    _embeddedWinId = id;
}

void EmbedFrame::unembed()
{
    if( _embeddedWinId )
        XReparentWindow(
            qt_xdisplay(), _embeddedWinId, qt_xrootwin(), 0, 0 );
}

WId EmbedFrame::embeddedWinId()
{
    return _embeddedWinId;
}

void EmbedFrame::resizeEvent( QResizeEvent *e )
{
    QWidget::resizeEvent( e );
}

void EmbedFrame::paintEvent( QPaintEvent *e )
{
    if( _bgPixmap->isNull() )
    {
        if( ! _iconPixmap->isNull() )
            bitBlt( this,
                (width() - _iconPixmap->width())/2,
                (height() - _iconPixmap->height())/2,
                _iconPixmap );
    }
    else
    {
        QPixmap *scaledBgPixmap = new QPixmap(
            Tools::getScaledPixmap(*_bgPixmap, QSize(64,64), true) );
        if( ! _iconPixmap->isNull() )
            bitBlt(scaledBgPixmap,
                (scaledBgPixmap->width() - _iconPixmap->width())/2,
                (scaledBgPixmap->height() - _iconPixmap->height())/2,
                _iconPixmap );
        bitBlt( this, 0, 0, scaledBgPixmap );
    }

    QWidget::paintEvent( e );
}

bool EmbedFrame::x11Event( XEvent *e )
{
    switch( e->type ) {
        case DestroyNotify:
            if( e->xdestroywindow.window == _embeddedWinId ) {
                _embeddedWinId = 0;
            }
            break;
        case ReparentNotify:
            if( _embeddedWinId &&
                (e->xreparent.window == _embeddedWinId) &&
                (e->xreparent.parent != winId()) ) {
                _embeddedWinId = 0;
            }
            else if( e->xreparent.parent == winId() ) {
                _embeddedWinId = e->xreparent.window;
                embed( _embeddedWinId );
            }
            break;
    }

    return false;
}


AppFrameBar::AppFrameBar( QWidget *parent, const char* name )
    : QWidget(parent,name)
{
}

void AppFrameBar::paintEvent( QPaintEvent *e )
{
    const QPixmap *bgPm = backgroundPixmap();
    if( (bgPm==0) || bgPm->isNull() )
    {
        QPainter painter;
        painter.begin(this);
        if( width() > height() )
        {
            for( int y=1+LINE_WIDTH; y<height()-1-LINE_WIDTH; y+=2*LINE_WIDTH )
            {
                qDrawShadeLine(&painter,
                    1+LINE_WIDTH, y, width()-1-LINE_WIDTH, y,
                    colorGroup(), false, LINE_WIDTH, 0);
            }
        }
        else
        {
            for( int x=1+LINE_WIDTH; x<width()-1-LINE_WIDTH; x+=2*LINE_WIDTH )
            {
                qDrawShadeLine(&painter,
                    x, 1+LINE_WIDTH, x, height()-1-LINE_WIDTH,
                    colorGroup(), false, LINE_WIDTH, 0);
            }
        }
        painter.end();
    }

    QWidget::paintEvent( e );
}


AppFrame::AppFrame( AdConfig *conf, KWinModule *module, Style *style,
    QStrList strList, QWidget *parent )
    : QFrame(parent, ""), config(conf), winModule(module), _style(style)
{
    if( strList.count() == 0 ) {
        appName = "";
        appCommand = "";
        appWindowTitle = "";
        appIcon = "";
        swallowApp = true;
        execOnStartup = true;
        closeOnExit = true;
        onlyStartOnce = true;
		if( conf->getDirection() == Common::HORIZONTAL )
	        heightOrWidth = conf->getFrameSize().width();
	    else
	    	heightOrWidth = conf->getFrameSize().height();	
    }
    else {
        appName = strList.at(0);
        appCommand = strList.at(1);
        appWindowTitle = strList.at(2);
        appIcon = strList.at(3);
        swallowApp = Tools::string2Bool( strList.at(4) );
        execOnStartup = Tools::string2Bool( strList.at(5) );
        closeOnExit = Tools::string2Bool( strList.at(6) );
        onlyStartOnce = Tools::string2Bool( strList.at(7) );
        heightOrWidth = QString(strList.at(8)).toInt();
    }

    QToolTip::add( this, appName );

    startTimer = new QTimer( this );
    QObject::connect( startTimer, SIGNAL(timeout()),
        SLOT(slotStartProcess()) );

    proc = 0;

    embedFrame = new EmbedFrame( winModule, this, "embedFrame" );
    embedFrame->installEventFilter( this );

    bar = new AppFrameBar( this, "bar" );
    bar->installEventFilter(this);

    dialog = new AppFrameDialogImpl(this, "appFrameDialog", true);

    applyPixmaps();
    resizeFromConfig();
    rearrange();

    initMenu();

    setFrameStyle( QFrame::Panel | QFrame::Raised );
    setLineWidth( LINE_WIDTH );
}

AppFrame::~AppFrame()
{
//kdDebug() << "AppFrame::~AppFrame()" << endl;
    if( closeOnExit ) {
        kill();
    }
    else {
        embedFrame->unembed();
    }
}

AppFrame *AppFrame::initWithDialog( AdConfig *config, KWinModule *module,
    Style *style, QWidget *parent )
{
    AppFrame *appFrame = new AppFrame(
        config, module, style, QStrList(), parent );
    if( appFrame->execDialog() ) {
        return appFrame;
    }
    else {
        delete appFrame;
        return 0;
    }
}

void AppFrame::initMenu()
{
    menu = new QPopupMenu();
    menu->insertItem(
        BarIcon("exec", KIcon::SizeSmall),
        i18n("Start"), Common::START );
    menu->insertItem(
        BarIcon("stop", KIcon::SizeSmall),
        i18n("Stop"), Common::STOP );
    menu->insertItem(
        BarIcon("pencil", KIcon::SizeSmall),
        i18n("Edit..."), Common::EDIT );
    menu->insertItem(
        BarIcon("fileclose", KIcon::SizeSmall),
        i18n("Delete"), Common::DELETE );
    menu->insertSeparator();
    menu->insertItem(
        BarIcon("filenew", KIcon::SizeSmall),
        i18n("New"), Common::NEW );
    menu->insertItem(
        BarIcon("configure", KIcon::SizeSmall),
        i18n("Preferences..."), Common::PREFS );
    menu->insertSeparator();
    menu->insertItem(
        BarIcon("help", KIcon::SizeSmall),
        i18n("Help"), Common::HELP );
    menu->insertItem(
        BarIcon("kappdock", KIcon::SizeSmall),
        i18n("About"), Common::ABOUT );
    menu->insertSeparator();
    menu->insertItem(
        BarIcon("exit", KIcon::SizeSmall),
        i18n("Exit"), Common::QUIT );
    QObject::connect( menu, SIGNAL(activated(int)),
        SLOT(slotCommandCallback(int)) );
}

void AppFrame::showMenu(const QPoint& p)
{
    if( embedFrame->embeddedWinId() )
        menu->setItemEnabled( Common::START, false );
    else
        menu->setItemEnabled( Common::START, true );
    if( (proc!=0) && proc->isRunning() )
        menu->setItemEnabled( Common::STOP, true );
    else
        menu->setItemEnabled( Common::STOP, false );
    menu->popup( p );
}

bool AppFrame::execDialog()
{
    QString oldCommand = appCommand;
    QString oldWindowTitle = appWindowTitle;
    QString oldIcon = appIcon;
    int oldHeightOrWidth = heightOrWidth;

    dialog->setName( appName );
    dialog->setCommand( appCommand );
    dialog->setWindowTitle( appWindowTitle );
    dialog->setIcon( appIcon );
    dialog->setSwallowApp( swallowApp );
    dialog->setExecOnStartup( execOnStartup );
    dialog->setCloseOnExit( closeOnExit );
    dialog->setOnlyStartOnce( onlyStartOnce );
    dialog->setHeightOrWidth( heightOrWidth );


    if( config->getDirection() == Common::HORIZONTAL )
	    dialog->setHeightOrWidthLabelText(i18n("Width"));
	else
		dialog->setHeightOrWidthLabelText(i18n("Height"));

    bool dialogExecuted = dialog->exec();
    if( dialogExecuted ) {
        appName = dialog->getName();
        appCommand = dialog->getCommand();
        appWindowTitle = dialog->getWindowTitle();
        appIcon = dialog->getIcon();
        swallowApp = dialog->getSwallowApp();
        execOnStartup = dialog->getExecOnStartup();
        closeOnExit = dialog->getCloseOnExit();
        onlyStartOnce = dialog->getOnlyStartOnce();
        heightOrWidth = dialog->getHeightOrWidth();

        if( appIcon != oldIcon )
            loadIconFromConfig();

        // only restart if command or window title has changed
        if( (oldCommand != appCommand) ||
            (oldWindowTitle != appWindowTitle) ) {
            if(closeOnExit || embedFrame->embeddedWinId())
                kill();

            if(execOnStartup)
                start();
        }

        if( heightOrWidth != oldHeightOrWidth )
        	resizeFromConfig();

        emit updateSignal( this );
    }

    return dialogExecuted;
}

void AppFrame::start( int latency )
{
    if( (proc != 0) && (onlyStartOnce) ) {
        if( proc->isRunning() ) proc->kill(SIGQUIT);
        delete proc;
        proc = 0;
    }

    if( (appCommand != 0) && (appCommand != "") ) {
        proc = new KProcess();
        if( onlyStartOnce )
            QObject::connect( proc, SIGNAL(processExited(KProcess*)),
                SLOT(slotProcessExited(KProcess*)) );

        // handle command strings with white spaces
        QString tmp=appCommand;
        int i=0;
        if( tmp.find(' ') != -1 )
            while((i=tmp.find(' ')) >=0 ) {
                *proc << tmp.left(i);
                tmp=tmp.right(tmp.length()-i-1);
            }
        *proc << tmp;

        if( swallowApp )
            winModule->doNotManage( appWindowTitle );
        startTimer->start(latency);
    }
}

void AppFrame::kill( int type )
{
    if( proc != 0 ) {
        if( proc->isRunning() ) {
            proc->kill(type);
        }
        delete proc;
        proc = 0;
    }
}

bool AppFrame::trySwallow( WId w )
{
    QString windowTitle = Tools::getWindowProperty(w, "WM_NAME");
    if( (windowTitle == appWindowTitle) &&
        (!embedFrame->embeddedWinId()) &&
        (swallowApp) ) {

        embedFrame->embed( w );

        repaint(false);
        return true;
    }
    return false;
}

bool AppFrame::tryUnswallow( WId w )
{
    if( embedFrame->embeddedWinId() == w ) {
        embedFrame->unembed();
        return true;
    }
    return false;
}

void AppFrame::loadIconFromConfig()
{
    QSize pmSize = embedFrame->size();
    QPixmap iconPm;
    if( appIcon != "" )
    {
        KIconLoader *ldr  = KGlobal::iconLoader();
        iconPm =
            ldr->loadIcon(appIcon, KIcon::Desktop, KIcon::SizeLarge);
    }

    if( ! iconPm.isNull () ) {
        iconPm = Tools::getScaledPixmap( iconPm, pmSize, false );
    }

    embedFrame->setIconPixmap( new QPixmap(iconPm) );
    embedFrame->repaint(false);
}

void AppFrame::resizeFromConfig()
{
    int w;
    int h;
    if( config->getDirection() == Common::HORIZONTAL )
    {
    	w = heightOrWidth;
    	h = config->getFrameSize().height();
    }
    else
    {
    	w = config->getFrameSize().width();
    	h = heightOrWidth;
    }

    if( (config->getFrameAlignment() == Common::TOP) ||
        (config->getFrameAlignment() == Common::BOTTOM) ) {
        bar->setFixedSize( w, BAR_WIDTH );
        embedFrame->setFixedSize( w, h );
        setFixedSize( w + 2*LINE_WIDTH, h + 2*LINE_WIDTH + BAR_WIDTH );
    }
    else if( (config->getFrameAlignment() == Common::RIGHT) ||
        (config->getFrameAlignment() == Common::LEFT) ) {
        bar->setFixedSize( BAR_WIDTH, h );
        embedFrame->setFixedSize( w, h );
        setFixedSize( w + 2*LINE_WIDTH + BAR_WIDTH, h + 2*LINE_WIDTH );
    }

    applyPixmaps();
}

void AppFrame::setStyle( Style *style )
{
//    kdDebug() << "set style" << endl;
    _style = style;
    applyPixmaps();
}

void AppFrame::applyPixmaps()
{
    if( _style == 0 ) return;

    embedFrame->setBgPixmap(_style->getAppframePixmap());
    if( (config->getFrameAlignment() == Common::TOP) ||
        (config->getFrameAlignment() == Common::BOTTOM) )
    {
        bar->setBackgroundPixmap(
            *_style->getAppframeHorizontalBarPixmap());
    }
    else if( (config->getFrameAlignment() == Common::RIGHT) ||
        (config->getFrameAlignment() == Common::LEFT) )
    {
        bar->setBackgroundPixmap(
            *_style->getAppframeVerticalBarPixmap());
    }
}

void AppFrame::setName( const QString& name )
{ appName = name; }

void AppFrame::setCommand( const QString& command )
{ appCommand = command; }

void AppFrame::setWindowTitle( const QString& windowTitle )
{ appWindowTitle = windowTitle; }

void AppFrame::setIcon( const QString& icon )
{ appIcon = icon; }

void AppFrame::setSwallowApp( bool enable )
{ swallowApp = enable; }

void AppFrame::setExecOnStartup( bool enable )
{ execOnStartup = enable; }

void AppFrame::setCloseOnExit( bool enable )
{ closeOnExit = enable; }

void AppFrame::setOnlyStartOnce( bool enable )
{ onlyStartOnce = enable; }

void AppFrame::setHeightOrWidth( int value )
{ heightOrWidth = value; }

// void AppFrame::setAlignment( Common::Alignment align )
// {
//     alignment = align;
//
//     rearrange();
// }

void AppFrame::rearrange()
{
    int w;
    int h;
    if( config->getDirection() == Common::HORIZONTAL )
    {
    	w = heightOrWidth;
    	h = config->getFrameSize().height();
    }
    else
    {
    	w = config->getFrameSize().width();
    	h = heightOrWidth;
    }

    if( config->getFrameAlignment() == Common::TOP ) {
        bar->move( LINE_WIDTH, LINE_WIDTH );
        embedFrame->move( LINE_WIDTH, LINE_WIDTH + BAR_WIDTH );
    }
    else if( config->getFrameAlignment() == Common::BOTTOM ) {
        bar->move( LINE_WIDTH, h + LINE_WIDTH );
        embedFrame->move( LINE_WIDTH, LINE_WIDTH );
    }
    else if( config->getFrameAlignment() == Common::LEFT ) {
        bar->move( LINE_WIDTH, LINE_WIDTH );
        embedFrame->move( LINE_WIDTH + BAR_WIDTH, LINE_WIDTH );
    }
    else if( config->getFrameAlignment() == Common::RIGHT ) {
        bar->move( w + LINE_WIDTH, LINE_WIDTH );
        embedFrame->move( LINE_WIDTH, LINE_WIDTH );
    }
}

QStrList *AppFrame::getStrList()
{
    QStrList *list = new QStrList();
    list->append( appName.latin1() );
    list->append( appCommand.latin1() );
    list->append( appWindowTitle.latin1() );
    list->append( appIcon.latin1() );
    list->append( Tools::bool2String(swallowApp).latin1() );
    list->append( Tools::bool2String(execOnStartup).latin1() );
    list->append( Tools::bool2String(closeOnExit).latin1() );
    list->append( Tools::bool2String(onlyStartOnce).latin1() );
    list->append( QString::number(heightOrWidth) );
    return list;
}

void AppFrame::resizeEvent( QResizeEvent *e )
{
    loadIconFromConfig();

    QFrame::resizeEvent( e );
}

bool AppFrame::eventFilter( QObject *ob, QEvent *ev )
{
    if( (ob == embedFrame) && !embedFrame->embeddedWinId() ) {
    // don't catch events if the embed frame has swallowed a window
        if( ev->type() == QEvent::MouseButtonDblClick ) {
            start();
            return true;
        }
        else if( ev->type() == QEvent::MouseButtonPress ) {
            QMouseEvent *me = static_cast<QMouseEvent*>(ev);
            if( me->button() == RightButton ){
                showMenu( me->globalPos() );
                return true;
            }
        }
    }
    else if( ob == bar ) {
        if( ev->type() == QEvent::MouseMove ) {
            emit mouseEventSignal( this, static_cast<QMouseEvent*>(ev) );
            return true;
        }
        else if( ev->type() == QEvent::MouseButtonPress ) {
            QMouseEvent *me = static_cast<QMouseEvent*>(ev);
            if( me->button() == RightButton ) {
                 showMenu( me->globalPos() );
                 return true;
            }
        }
    }

    return false;
}

// void AppFrame::mousePressEvent( QMouseEvent *e )
// {
//     if( e->button() == RightButton )
//         menu->popup( e->globalPos() );
// }

void AppFrame::slotStartProcess()
{
    startTimer->stop();

//    winModule->doNotManage( appWindowTitle );

    if( onlyStartOnce )
        proc->start( KProcess::NotifyOnExit, KProcess::NoCommunication );
    else
        proc->start( KProcess::DontCare, KProcess::NoCommunication );

    repaint(false);
}

void AppFrame::slotProcessExited(KProcess *)
{
}

void AppFrame::slotCommandCallback( int id )
{
    switch(id) {
        case Common::START:
            if( !embedFrame->embeddedWinId() )
                start();
            break;
        case Common::STOP:
            if( (proc!=0) && proc->isRunning() )
                kill();
            break;
        case Common::EDIT:
            execDialog();
            break;
        default:
            emit menuEventSignal(this, (Common::MenuId) id);
            break;
    }
}

#include "appframe.moc"
