#include "toplevel.h"
#include "adconfig.h"
#include "appframe.h"
#include "parser.h"
#include "prefsdialogimpl.h"

#include <qstrlist.h>
#include <qstyle.h>
#include <qpainter.h>
#include <qcolor.h>
#include <qbitmap.h>
#include <qpixmap.h>
#include <qdrawutil.h>
#include <qapplication.h>
#include <qmessagebox.h>
#include <qvaluelist.h>
#include <qregexp.h>
#include <qstringlist.h>
#include <qtextstream.h>

#include <kwinmodule.h>
#include <kwin.h>
#include <kiconloader.h>
#include <klocale.h>
#include <kdebug.h>
#include <kstddirs.h>
#include <ktar.h>
#include <kaboutdata.h>
#include <kapp.h>
#include <kpopupmenu.h>


ToplevelButton::ToplevelButton( QWidget *parent, const char *name )
	: QFrame( parent, name )
{
    setFrameStyle( QFrame::Panel | QFrame::Raised );
    setLineWidth( LINE_WIDTH );
}

void ToplevelButton::mousePressEvent( QMouseEvent *me )
{
	if( me->button() == QMouseEvent::LeftButton )
	{
        emit clickedSignal();
	}
}

void ToplevelButton::paintEvent( QPaintEvent *e )
{
    const QPixmap *bgPm = backgroundPixmap();
    if( (bgPm==0) || bgPm->isNull() )
    {
        QPainter painter;
        painter.begin(this);
#if QT_VERSION < 300
        qApp->style().drawArrow(
            &painter, Qt::DownArrow, false,
            LINE_WIDTH, LINE_WIDTH,
            width()-2*LINE_WIDTH, height()-2*LINE_WIDTH,
            colorGroup(), true );
#else
        qApp->style().drawPrimitive( QStyle::PE_ArrowDown, &painter, 
		QRect( LINE_WIDTH, LINE_WIDTH, width()-2*LINE_WIDTH, height()-2*LINE_WIDTH),
                colorGroup(), QStyle::Style_Enabled);
#endif
        painter.end();
    }

    QFrame::paintEvent(e);
}

ToplevelBar::ToplevelBar( QWidget *parent, const char *name )
    : QFrame( parent, name )
{
    setFrameStyle( QFrame::Panel | QFrame::Raised );
    setLineWidth( LINE_WIDTH );
}

void ToplevelBar::paintEvent( QPaintEvent *e )
{
    const QPixmap *bgPm = backgroundPixmap();
    if( (bgPm==0) || bgPm->isNull() )
    {
        QPainter painter;
        painter.begin(this);
        if( width() > height() )
        {
            for( int y=1+LINE_WIDTH; y<height()-1-LINE_WIDTH; y+=2*LINE_WIDTH )
            {
                qDrawShadeLine(&painter,
                    1+LINE_WIDTH, y, width()-1-LINE_WIDTH, y,
                    colorGroup(), false, LINE_WIDTH, 0);
            }
        }
        else
        {
            for( int x=1+LINE_WIDTH; x<width()-1-LINE_WIDTH; x+=2*LINE_WIDTH )
            {
                qDrawShadeLine(&painter,
                    x, 1+LINE_WIDTH, x, height()-1-LINE_WIDTH,
                    colorGroup(), false, LINE_WIDTH, 0);
            }
        }
        painter.end();
    }

    QFrame::paintEvent( e );
}


Style::Style( QPixmap *afPm, QPixmap *afVBPm, QPixmap *afHBPm,
	QPixmap *tlVBPm, QPixmap *tlHBPm, QPixmap *tlBPm )
	: _appframePixmap(afPm),
	  _appframeVerticalBarPixmap(afVBPm),
	  _appframeHorizontalBarPixmap(afHBPm),
	  _toplevelVerticalBarPixmap(tlVBPm),
	  _toplevelHorizontalBarPixmap(tlHBPm),
	  _toplevelButtonPixmap(tlBPm)
{
}

Style::Style()
{
    _appframePixmap = new QPixmap();
    _appframeVerticalBarPixmap = new QPixmap();
    _appframeHorizontalBarPixmap = new QPixmap();
    _toplevelVerticalBarPixmap = new QPixmap();
    _toplevelHorizontalBarPixmap = new QPixmap();
    _toplevelButtonPixmap = new QPixmap();
}

Style::~Style()
{
    if( _appframePixmap != 0) delete _appframePixmap;
    if( _appframeVerticalBarPixmap != 0) delete _appframeVerticalBarPixmap;
    if( _appframeHorizontalBarPixmap != 0) delete _appframeHorizontalBarPixmap;
    if( _toplevelVerticalBarPixmap != 0) delete _toplevelVerticalBarPixmap;
    if( _toplevelHorizontalBarPixmap != 0) delete _toplevelHorizontalBarPixmap;
    if( _toplevelButtonPixmap != 0) delete _toplevelButtonPixmap;
}

Toplevel::Toplevel( AdConfig *config, KAboutData *about )
    : KMainWindow(), _config(config), _about(about)
{
    _winModule = new KWinModule(0);
    QObject::connect( _winModule, SIGNAL(windowAdded(WId)),
        SLOT(slotWindowAdded(WId)) );
    QObject::connect( _winModule, SIGNAL(workAreaChanged()),
        SLOT(slotWorkAreaChanged()) );

    _prefsDlg = new PrefsDialogImpl(this, "prefsDlg", false);
    QObject::connect( _prefsDlg, SIGNAL(okButtonClickedSignal()),
        SLOT(slotPrefsDlgOk()) );
    QObject::connect( _prefsDlg, SIGNAL(applyButtonClickedSignal()),
        SLOT(slotPrefsDlgApply()) );
    QObject::connect( _prefsDlg, SIGNAL(cancelButtonClickedSignal()),
        SLOT(slotPrefsDlgCancel()) );

    _style = loadStyle( _config->getStyleName() );

    appFrameWidget = new QWidget( this, "appFrameWidget" );

    initMenu();

    menubar = new ToplevelBar( this, "menubar" );
    menubar->installEventFilter(this);
    menubar->show();

	minimizeButton = new ToplevelButton( this, "minimizeButton" );
	QObject::connect( minimizeButton, SIGNAL(clickedSignal()),
		SLOT(slotMinimize()) );
	minimizeButton->show();

    QList<QStrList> applist = _config->getAppList();
    for( unsigned int i=0; i<applist.count(); i++ ) {
        addAppFrame( *(applist.at(i)) );
    }

    if( appFrameList.count() == 0 ) {
        QStrList list;
        list.append("");
        list.append("");
        list.append("");
        list.append("");
        list.append("true");
        list.append("true");
        list.append("true");
        list.append("true");
        if( _config->getDirection() == Common::HORIZONTAL )
	        list.append(QString::number(_config->getFrameSize().width()));
	    else
	    	list.append(QString::number(_config->getFrameSize().height()));

        addAppFrame( list );
    }

    // init system tray
    QString normalPixmap = KGlobal::dirs()->findResource(
        "ad_icons", "kappdock-normal.png" );
    QString darkerPixmap = KGlobal::dirs()->findResource(
        "ad_icons", "kappdock-darker.png" );
    QPixmap *normalPm = new QPixmap(normalPixmap);
    QPixmap *darkerPm = new QPixmap(darkerPixmap);

    _sysTray = new AdSysTray( normalPm, darkerPm, this, "_sysTray" );
    QObject::connect( _sysTray, SIGNAL(mouseEventSignal(QMouseEvent *)),
        this, SLOT(slotSysTrayMouseEvent(QMouseEvent *)) );
	if( config->isSystemTrayVisible() )
	    _sysTray->show();

    KWin::setType( winId(), NET::Dock );
    KWin::setState( winId(), NET::Sticky | NET::StaysOnTop );
    KWin::setOnAllDesktops( winId(), true );

    resizeAppFramesFromConfig();
    rearrangeAppFrames();
    rearrange();
    reposition();
    applyStyle();
    show();

    updateSysTray();

    // start applications
    for( int unsigned i=0; i < appFrameList.count(); i++ )
        if( appFrameList.at(i)->getExecOnStartup() )
            appFrameList.at(i)->start(5000);
}

Toplevel::~Toplevel() {
}

Style *Toplevel::loadStyle(const QString& styleName)
{
//	kdDebug() << styleName << endl;
	QString styleFile = KGlobal::dirs()->findResource("ad_styles",
		styleName + ".style");		
	KTarGz *tarGz = new KTarGz(styleFile);
	tarGz->open(IO_ReadOnly);
	if( ! tarGz->isOpened() )
	{
        kdDebug() << "The file \"" << styleFile <<
            "\" could not be opened." << endl;
        return new Style();
    }
	
	const KTarDirectory *tarDir = tarGz->directory();
	const KTarEntry *tarEntry = tarDir->entry(styleName + ".stylerc");
	if( (tarEntry == 0) || (! tarEntry->isFile()) )
	{
	    kdDebug() << "The file \"" << styleName <<
	        ".stylerc\" was not found or could not be opened." << endl;
	    return new Style();
	}

    const KTarFile *tarFile = static_cast<const KTarFile*>(tarEntry);
    QTextStream stream(tarFile->data(),IO_ReadOnly);
    AdSimpleParser parser( &stream );
    parser.parse();

    QString pmString;
    const KTarEntry *entry;
    QPixmap
        *appframePixmap = new QPixmap(),
        *appframeVerticalBarPixmap = new QPixmap(),
        *appframeHorizontalBarPixmap = new QPixmap(),
        *toplevelVerticalBarPixmap = new QPixmap(),
        *toplevelHorizontalBarPixmap = new QPixmap(),
        *toplevelButtonPixmap = new QPixmap();

    pmString = parser.getValue( "pixmaps", "appframe" );
    if( pmString != QString::null ) {
        entry = tarDir->entry(pmString);
        if( (entry != 0) && entry->isFile() ) {
            const KTarFile *fileEntry = static_cast<const KTarFile*>(entry);
            appframePixmap = new QPixmap(fileEntry->data());
        }
    }

    pmString = parser.getValue( "pixmaps", "appframeVerticalBar" );
    if( pmString != QString::null ) {
        entry = tarDir->entry(pmString);
        if( (entry != 0) && entry->isFile() ) {
            const KTarFile *fileEntry = static_cast<const KTarFile*>(entry);
            appframeVerticalBarPixmap = new QPixmap(fileEntry->data());
        }
    }

    pmString = parser.getValue( "pixmaps", "appframeHorizontalBar" );
    if( pmString != QString::null ) {
        entry = tarDir->entry(pmString);
        if( (entry != 0) && entry->isFile() ) {
            const KTarFile *fileEntry = static_cast<const KTarFile*>(entry);
            appframeHorizontalBarPixmap = new QPixmap(fileEntry->data());
        }
    }

    pmString = parser.getValue( "pixmaps", "toplevelVerticalBar" );
    if( pmString != QString::null ) {
        entry = tarDir->entry(pmString);
        if( (entry != 0) && entry->isFile() ) {
            const KTarFile *fileEntry = static_cast<const KTarFile*>(entry);
            toplevelVerticalBarPixmap = new QPixmap(fileEntry->data());
        }
    }

    pmString = parser.getValue( "pixmaps", "toplevelHorizontalBar" );
    if( pmString != QString::null ) {
        entry = tarDir->entry(pmString);
        if( (entry != 0) && entry->isFile() ) {
            const KTarFile *fileEntry = static_cast<const KTarFile*>(entry);
            toplevelHorizontalBarPixmap = new QPixmap(fileEntry->data());
        }
    }

    pmString = parser.getValue( "pixmaps", "toplevelButton" );
    if( pmString != QString::null ) {
        entry = tarDir->entry(pmString);
        if( (entry != 0) && entry->isFile() ) {
            const KTarFile *fileEntry = static_cast<const KTarFile*>(entry);
            toplevelButtonPixmap = new QPixmap(fileEntry->data());
        }
    }

    return new Style(
        appframePixmap,
        appframeVerticalBarPixmap, appframeHorizontalBarPixmap,
        toplevelVerticalBarPixmap, toplevelHorizontalBarPixmap,
        toplevelButtonPixmap );
}

void Toplevel::initMenu()
{
    menu = new KPopupMenu();
    menu->insertItem(
        BarIcon("filenew", KIcon::SizeSmall),
        i18n("New"), Common::NEW );
    menu->insertItem(
        BarIcon("configure", KIcon::SizeSmall),
        i18n("Preferences..."), Common::PREFS );
    menu->insertSeparator();
    menu->insertItem(
        BarIcon("help", KIcon::SizeSmall),
        i18n("Help"), Common::HELP );
    menu->insertItem(
        BarIcon("kappdock", KIcon::SizeSmall),
        i18n("About"), Common::ABOUT );
    menu->insertSeparator();
    menu->insertItem(
        BarIcon("exit", KIcon::SizeSmall),
        i18n("Exit"), Common::QUIT );
    QObject::connect( menu, SIGNAL(activated(int)),
        SLOT(slotCommandCallback(int)) );
}

void Toplevel::applyStyle()
{
//kdDebug() << "Toplevel::applyStyle" << endl;
    for( unsigned int i=0; i<appFrameList.count(); i++ )
        appFrameList.at(i)->setStyle(_style);
    applyPixmaps();
}

void Toplevel::applyPixmaps()
{
//kdDebug() << "Toplevel::applyPixmaps" << endl;
    minimizeButton->setBackgroundPixmap(
        *_style->getToplevelButtonPixmap() );

    if( _config->getDirection() == Common::HORIZONTAL )
        menubar->setBackgroundPixmap(
            *_style->getToplevelVerticalBarPixmap() );
    else
        menubar->setBackgroundPixmap(
            *_style->getToplevelHorizontalBarPixmap() );
}

void Toplevel::updateSysTray()
{
    if( isVisible() )
        _sysTray->setPixmap( AdSysTray::PmNormal );
    else
        _sysTray->setPixmap( AdSysTray::PmDarker );

    _sysTray->repaint();
}

void Toplevel::slotCommandCallback( int id )
{
    menuEvent((Common::MenuId) id);
}

void Toplevel::setStyle( Style *style )
{
	if( _style != 0 ) delete _style;
	_style = style;
	applyStyle();
}

void Toplevel::addAppFrame( AppFrame *frame )
{
    if( frame != 0 ) {
        appFrameList.append(frame);
        frame->show();
        QObject::connect(
            frame, SIGNAL(mouseEventSignal(QWidget*,QMouseEvent*)),
            SLOT(slotAppFrameMouseEvent(QWidget*,QMouseEvent*)) );
        QObject::connect(
            frame, SIGNAL(menuEventSignal(AppFrame*,Common::MenuId)),
            SLOT(slotAppFrameMenuEvent(AppFrame*,Common::MenuId)) );
        QObject::connect(
            frame, SIGNAL(updateSignal(AppFrame*)),
            SLOT(slotAppFrameUpdated(AppFrame*)) );
    }
    rearrange();
}

void Toplevel::addAppFrame( QStrList list )
{
    AppFrame *frame =
        new AppFrame( _config, _winModule, _style, list, appFrameWidget );
    addAppFrame( frame );
}

void Toplevel::resizeAppFramesFromConfig()
{
    for( unsigned int i=0; i<appFrameList.count(); i++ )
        appFrameList.at(i)->resizeFromConfig();

    rearrange();
}

void Toplevel::rearrangeAppFrames()
{
    for( unsigned int i=0; i<appFrameList.count(); i++ )
    {
        appFrameList.at(i)->resizeFromConfig();
        appFrameList.at(i)->rearrange();
    }
}

bool Toplevel::trySwallow(WId id)
{
    for( unsigned int i=0; i<appFrameList.count(); i++ )
        if( appFrameList.at(i)->trySwallow(id) )
            return true;

    return false;
}

bool Toplevel::tryUnswallow(WId id)
{
    for( unsigned int i=0; i<appFrameList.count(); i++ )
        if( appFrameList.at(i)->tryUnswallow(id) )
            return true;

    return false;
}

/**
 * Rearrange app windows within the toplevel window and
 * reposition menu bar.
 */
void Toplevel::rearrange()
{
    if( _config->getDirection() == Common::VERTICAL ) {
        int width = appFrameList.at(0)->width();
        int height = 0;
        for( unsigned int i=0; i<appFrameList.count(); i++ ) {
            appFrameList.at(i)->move( 0, height );
            height += appFrameList.at(i)->height();
		}

		minimizeButton->setFixedSize( MENUBAR_WIDTH, MENUBAR_WIDTH );
		minimizeButton->move( width-MENUBAR_WIDTH, 0 );
		minimizeButton->show();
        menubar->setFixedSize( width-MENUBAR_WIDTH, MENUBAR_WIDTH );
        menubar->move( 0, 0 );
        menubar->show();
        appFrameWidget->setFixedSize( width, height );
        appFrameWidget->move( 0, menubar->height() );
        setFixedSize( width, height + menubar->height() );
    }
    else if( _config->getDirection() == Common::HORIZONTAL ) {
        int height = appFrameList.at(0)->height();
        int width = 0;
        for( unsigned int i=0; i<appFrameList.count(); i++ ) {
            appFrameList.at(i)->move( width, 0 );
            width += appFrameList.at(i)->width();
        }

		minimizeButton->setFixedSize( MENUBAR_WIDTH, MENUBAR_WIDTH );
		minimizeButton->move( 0, 0 );
		minimizeButton->show();
        menubar->setFixedSize( MENUBAR_WIDTH, height-MENUBAR_WIDTH );
        menubar->move( 0, MENUBAR_WIDTH );
        menubar->show();
        appFrameWidget->setFixedSize( width, height );
        appFrameWidget->move( menubar->width(), 0 );
        setFixedSize( width + menubar->width(), height );
    }

    applyPixmaps();
}

/**
 * Position toplevel window on desktop.
 */
void Toplevel::reposition()
{
    QRect wA = _winModule->workArea();
    QRect geom = geometry();
    QPoint confPos = _config->getGeometry().topLeft();
    QPoint moveTo;
    int snapZone = _config->getBorderSnapZone();

    if( _config->getPosition() & Common::POS_LEFT )
    {
        moveTo.setX(wA.left());
    }
    else if( _config->getPosition() & Common::POS_RIGHT )
    {
        moveTo.setX(wA.right()-geom.width());
    }
    else
    {
        if( (confPos.x() < wA.left()+snapZone)
        	&& (confPos.x() > wA.left()-snapZone) )
            moveTo.setX(wA.left());
        else if( (confPos.x() > wA.right()-geom.width()-snapZone)
        	&& (confPos.x() < wA.right()-geom.width()+snapZone) )
            moveTo.setX(wA.right()-geom.width());
        else
            moveTo.setX(confPos.x());
    }

    if( _config->getPosition() & Common::POS_TOP )
    {
        moveTo.setY(wA.top());
    }
    else if( _config->getPosition() & Common::POS_BOTTOM )
    {
        moveTo.setY(wA.bottom()-geom.height());
    }
    else
    {
        if( (confPos.y() < wA.top()+snapZone)
        	&& (confPos.y() > wA.top()-snapZone) )
            moveTo.setY(wA.top());
        else if( (confPos.y() > wA.bottom()-geom.height()-snapZone)
        	&& (confPos.y() < wA.bottom()-geom.height()+snapZone) )
            moveTo.setY(wA.bottom()-geom.height());
        else
            moveTo.setY(confPos.y());
    }

    move( moveTo );
    _config->setGeometry(geometry());
}

void Toplevel::updateConfig()
{
    QList<QStrList> appList = _config->getAppList();
    appList.clear();

    for( unsigned int i=0; i<appFrameList.count(); i++ )
        appList.append( appFrameList.at(i)->getStrList() );

    _config->setAppList( appList );
    _config->setAppCount( appFrameList.count() );
    _config->setGeometry( geometry() );
}

bool Toplevel::eventFilter( QObject *ob, QEvent *ev )
{
    if( ob == menubar )
    {
        static QRect wA = QRect(0,0,0,0);
        static QPoint pos1 = QPoint(0,0);
        static QRect geom = QRect(0,0,0,0);

        if( ev->type() == QEvent::MouseButtonPress )
        {
            QMouseEvent *me = static_cast<QMouseEvent*>(ev);
            if( me->button() == RightButton )
            {
                menu->popup( me->globalPos() );
            }
            else if( me->button() == LeftButton )
            {
                wA = _winModule->workArea();
                pos1 = mapFromGlobal(me->globalPos());
                geom = geometry();
            }
            return true;
        }
        else if( ev->type() == QEvent::MouseMove )
        {
            QMouseEvent *me = static_cast<QMouseEvent*>(ev);
            int pos = _config->getPosition();
            int moveToX, moveToY;
            int snapZone = _config->getBorderSnapZone();
            if( pos & Common::POS_LEFT )
                moveToX = wA.left();
            else if( pos & Common::POS_RIGHT )
                moveToX = wA.right()-geom.width();
            else
                moveToX = (me->globalPos() - pos1).x();

            if( pos & Common::POS_TOP )
                moveToY = wA.top();
            else if( pos & Common::POS_BOTTOM )
                moveToY = wA.bottom()-geom.height();
            else
                moveToY = (me->globalPos() - pos1).y();

            if( (moveToX < wA.left()+snapZone)
            	&& (moveToX > wA.left()-snapZone) )
                moveToX = wA.left();
            else if( (moveToX > wA.right()-geom.width()-snapZone)
            	&& (moveToX < wA.right()-geom.width()+snapZone) )
                moveToX = wA.right()-geom.width();

            if( (moveToY < wA.top()+snapZone)
            	&& (moveToY > wA.top()-snapZone) )
                moveToY = wA.top();
            else if( (moveToY > wA.bottom()-geom.height()-snapZone)
            	&& (moveToY < wA.bottom()-geom.height()+snapZone) )
                moveToY = wA.bottom()-geom.height();

            move( moveToX, moveToY );
            return true;
        }
        else if( ev->type() == QEvent::MouseButtonRelease )
        {
            QMouseEvent *me = static_cast<QMouseEvent*>(ev);
            if( me->button() == LeftButton )
            {
                _config->setGeometry(geometry());
            }
        }
    }

    return false;
}

bool Toplevel::queryExit()
{
//kdDebug() << "Toplevel::queryExit" << endl;
    save();

    for( unsigned int i=0; i<appFrameList.count(); i++ )
    {
        AppFrame *appFrame = appFrameList.at(i);
        if( appFrame->getCloseOnExit() )
            appFrame->kill();
    }

    return true;
}

void Toplevel::createAppFrame()
{
    AppFrame *frame = AppFrame::initWithDialog(
        _config, _winModule, _style, appFrameWidget );
    if( frame != 0 ) {
        addAppFrame( frame );
        rearrange();
        reposition();
        save();
        if( frame->getExecOnStartup() )
            frame->start(100);
    }
}

void Toplevel::menuEvent(Common::MenuId id)
{
    switch( id ) {
        case Common::NEW: createAppFrame(); break;
        case Common::PREFS: preferences(); break;
        case Common::QUIT: quit(); break;
        case Common::HELP: help(); break;
        case Common::ABOUT: about(); break;
        default: break;
    }
}

void Toplevel::about()
{
    QString text = "";
    text += _about->programName() + " ";
    text += _about->version() + "\n";
    text += "Author(s):\n";
    QValueList<KAboutPerson> list = _about->authors();
    for( uint i=0; i<list.count(); i++ )
        text += list[i].name() + " <" + list[i].emailAddress() + ">\n";

    QMessageBox::information(this, _about->programName(), text);
}

void Toplevel::help()
{
    KApplication::kApplication()->invokeHelp();
}

void Toplevel::preferences()
{
    static QRegExp fileRegExp("[^/]+.style$");

    _prefsDlg->setDirection( _config->getDirection() );
    _prefsDlg->setPosition( _config->getPosition() );
    _prefsDlg->setFrameAlignment( _config->getFrameAlignment() );
    _prefsDlg->setFrameSize( _config->getFrameSize() );
    QStringList
        fileList,
        longFileList = KGlobal::dirs()->findAllResources(
            "ad_styles", "*.style");
    QStringList::Iterator it;
    for( it=longFileList.begin(); it!=longFileList.end(); ++it)
    {
        QString s = *it;
        int len, pos;
        pos = fileRegExp.match(s,0,&len);
        s = s.mid(pos,len);
        pos = s.find(".style");
        s.truncate(pos);
//kdDebug() << s << endl;
        fileList.append(s);
    }
    _prefsDlg->setStyleNames(fileList);
    _prefsDlg->setCurrentStyleName(_config->getStyleName());
    _prefsDlg->setSystemTrayVisible(_config->isSystemTrayVisible());

    _prefsDlg->show();
}

void Toplevel::save()
{
    updateConfig();
    _config->writeSettings();
}

void Toplevel::quit()
{
    KApplication::kApplication()->quit();
}

void Toplevel::slotWindowAdded(WId id)
{
    trySwallow(id);
}

void Toplevel::slotWorkAreaChanged()
{
    reposition();
}

/**
 * Handle dragging app windows within toplevel window.
 */
void Toplevel::slotAppFrameMouseEvent( QWidget *w, QMouseEvent *e )
{
    if( e->type() == QEvent::MouseMove ) {
        int
            index1=0,
            index2=0;
        int count = appFrameList.count();
        QPoint pos = appFrameWidget->mapFromGlobal( e->globalPos() );

        if( _config->getDirection() == Common::VERTICAL ) {
            if( (pos.y() < 0) || (pos.y() > appFrameWidget->height()) )
                return;
            index1 = getAppFrameIndexByPos(w->y(), false);
            index2 = getAppFrameIndexByPos(pos.y(), false);
        }
        else if( _config->getDirection() == Common::HORIZONTAL ) {
            if( (pos.x() < 0) || (pos.x() > appFrameWidget->width()) )
                return;
            index1 = getAppFrameIndexByPos(w->x(), true);
            index2 = getAppFrameIndexByPos(pos.x(), true);
        }

        if( (index1 < 0) || (index1 >= count) ||
            (index2 < 0) || (index2 >= count) )
            return;

        if( index1 != index2 ) {
            AppFrame *frame = appFrameList.at(index1);
            appFrameList.remove(index1);
            appFrameList.insert(index2, frame);
            rearrange();
        }
    }
}

void Toplevel::slotAppFrameMenuEvent(AppFrame* frame, Common::MenuId id )
{
    // don't delete last appframe
    switch( id ) {
        case Common::DELETE:
            if( appFrameList.count() == 1 ) return;
            if( appFrameList.remove( frame ) ) {
                delete frame;
                rearrange();
                updateConfig();
                _config->writeSettings();
            }
            break;
        default:
            menuEvent(id); break;
    }
}

void Toplevel::slotAppFrameUpdated( AppFrame * )
{
    save();
    rearrange();
}

void Toplevel::slotMinimize()
{
	if( _config->isSystemTrayVisible() )
	{
		hide();
		updateSysTray();
	}
}

void Toplevel::slotPrefsDlgOk()
{
    QSize oldFrameSize = _config->getFrameSize();
    Common::Direction oldDirection = _config->getDirection();
    int oldPosition = _config->getPosition();
    Common::Alignment oldAlignment = _config->getFrameAlignment();
    QString oldStyle = _config->getStyleName();
    bool oldSystemTrayVisible = _config->isSystemTrayVisible();

    _config->setDirection(_prefsDlg->getDirection());
    _config->setPosition(_prefsDlg->getPosition());
    _config->setFrameSize(_prefsDlg->getFrameSize());
    _config->setFrameAlignment(_prefsDlg->getFrameAlignment());
    _config->setStyleName(_prefsDlg->getStyleName());
    _config->setSystemTrayVisible(_prefsDlg->isSystemTrayVisible());

    if( (_prefsDlg->getFrameSize() != oldFrameSize) ||
        (_prefsDlg->getFrameAlignment() != oldAlignment) ) {
        resizeAppFramesFromConfig();
        rearrangeAppFrames();
        rearrange();
        reposition();
    }

    if( (_prefsDlg->getDirection() != oldDirection) ||
        (_prefsDlg->getPosition() != oldPosition) )
    {
        rearrangeAppFrames();
        rearrange();
        reposition();
    }

    if( _prefsDlg->getStyleName() != oldStyle )
    {
        Style *style = loadStyle(_prefsDlg->getStyleName());
        setStyle(style);
    }

    if( _prefsDlg->isSystemTrayVisible() != oldSystemTrayVisible )
    {
		if( _config->isSystemTrayVisible() )
		{
			updateSysTray();
			_sysTray->show();
		}
		else
		{
			if( ! isVisible() )
				show();
			_sysTray->hide();
		}
    }

    save();
    _prefsDlg->hide();
}

void Toplevel::slotPrefsDlgApply()
{
    QSize oldFrameSize = _config->getFrameSize();
    Common::Direction oldDirection = _config->getDirection();
    int oldPosition = _config->getPosition();
    Common::Alignment oldAlignment = _config->getFrameAlignment();
    QString oldStyle = _config->getStyleName();
    bool oldSystemTrayVisible = _config->isSystemTrayVisible();

    _config->setDirection(_prefsDlg->getDirection());
    _config->setPosition(_prefsDlg->getPosition());
    _config->setFrameSize(_prefsDlg->getFrameSize());
    _config->setFrameAlignment(_prefsDlg->getFrameAlignment());
    _config->setStyleName(_prefsDlg->getStyleName());
    _config->setSystemTrayVisible(_prefsDlg->isSystemTrayVisible());

    if( (_prefsDlg->getFrameSize() != oldFrameSize) ||
        (_prefsDlg->getFrameAlignment() != oldAlignment) ) {
        resizeAppFramesFromConfig();
        rearrangeAppFrames();
        rearrange();
        reposition();
    }

    if( (_prefsDlg->getDirection() != oldDirection) ||
        (_prefsDlg->getPosition() != oldPosition) )
    {
        rearrangeAppFrames();
        rearrange();
        reposition();
    }

    if( _prefsDlg->getStyleName() != oldStyle )
    {
        Style *style = loadStyle(_prefsDlg->getStyleName());
        setStyle(style);
    }

    if( _prefsDlg->isSystemTrayVisible() != oldSystemTrayVisible )
    {
		if( _config->isSystemTrayVisible() )
		{
			updateSysTray();
			_sysTray->show();
		}
		else
		{
			if( ! isVisible() )
				show();
			_sysTray->hide();
		}
    }

    save();
}

void Toplevel::slotPrefsDlgCancel() {
    _prefsDlg->hide();
}

void Toplevel::slotSysTrayMouseEvent( QMouseEvent *e )
{
    if( e->button() == RightButton )
        menu->popup(e->globalPos());
    else if( e->button() == LeftButton )
        updateSysTray();

}

int Toplevel::getAppFrameIndexByPos( int pos, bool xAxis )
{
	if( xAxis )
	{
		for( unsigned int i=0; i<appFrameList.count(); i++ )
		{
			AppFrame *frame = appFrameList.at(i);
			if( frame->x() + frame->width() > pos )
				return i;
		}
	}
	else
	{
		for( unsigned int i=0; i<appFrameList.count(); i++ )
		{
			AppFrame *frame = appFrameList.at(i);
			if( frame->y() + frame->height() > pos )
				return i;
		}
	}
	return -1;
}


AdSysTray::AdSysTray( QPixmap *normalPm, QPixmap *darkerPm,
    QWidget *parent, const char *name )
    : KSystemTray( parent, name ), _normalPm(normalPm), _darkerPm(darkerPm)
{
}

void AdSysTray::setPixmap( AdSysTray::Pixmap pm )
{
    _pm = pm;
}

void AdSysTray::mousePressEvent( QMouseEvent *e )
{
    if( e->button() == RightButton )
    {
        emit mouseEventSignal( e );
    }
    else if( e->button() == LeftButton )
    {
        KSystemTray::mousePressEvent( e );
        emit mouseEventSignal( e );
    }
    else
    {
        KSystemTray::mousePressEvent( e );
    }
}

void AdSysTray::paintEvent( QPaintEvent *e )
{
    int x = (width()-_normalPm->width())/2;
    int y = (height()-_normalPm->height())/2;
    if( x < 0 ) x = 0;
    if( y < 0 ) y = 0;
    QPainter p;
    p.begin(this);
    switch( _pm )
    {
        case AdSysTray::PmNormal:
            p.drawPixmap( x, y, *_normalPm );
                break;
        case AdSysTray::PmDarker:
            p.drawPixmap( x, y, *_darkerPm );
                break;
        default: break;
    }
    p.end();

    KSystemTray::paintEvent( e );
}

#include "toplevel.moc"
